import { system, world } from "@minecraft/server";
import { log } from "../../utils/utils";
import TimerUtil from "../../utils/cooldowns";
import { getAbilityName } from "../../utils/addonUtils";

export const fell_tree_timer = new TimerUtil();

export const fell_tree_ability = {
  key: "fell_tree",
  unlock_level: 30,
  enableTitle: true,
  description: (level) => {
    return [`${getDuration(level)}`];
  },
  cooldown: (level) => getCooldown(level),
  onEnable: (data) => {
    const { player, skill, level } = data;
    fell_tree_timer.set(player.name, getDuration(level) * 20);
    fell_tree_timer.enableDisplay(player, () => {
      return {
        rawtext: [
          { text: "§p" },
          ...getAbilityName(fell_tree_ability),
          {
            text: ` ${(
              (fell_tree_timer.get(player.name) - system.currentTick) /
              20
            ).toFixed(1)}s`,
          },
        ],
      };
    });
    skill.abilityCooldown.set(player, getCooldown(level) * 20, skill);
  },
};

const getDuration = (level) => {
  let value = 5;
  if (level >= 2) value += 5;
  if (level >= 4) value += 5;
  return value;
};

const getCooldown = (level) => {
  let cooldown = 90;
  if (level >= 1) cooldown -= 5;
  if (level >= 3) cooldown -= 5;
  if (level >= 5) cooldown -= 10;
  return cooldown;
};

const FellingRecursionLimit = {
  h: 8,
  v: 48,
};

const logs = [
  "acacia_log",
  "birch_log",
  "cherry_log",
  "dark_oak_log",
  "jungle_log",
  "mangrove_log",
  "oak_log",
  "spruce_log",
];

async function fellBlock(
  bannedDirections,
  block,
  sourceBlock,
  locationObject = {}
) {
  if (!block || !block.isValid() || (block.isAir && sourceBlock !== block))
    return;
  if (
    Math.abs(block.x - sourceBlock.x) > FellingRecursionLimit.h ||
    Math.abs(block.z - sourceBlock.z) > FellingRecursionLimit.h ||
    Math.abs(block.y - sourceBlock.y) > FellingRecursionLimit.v
  )
    return;

  if (block !== sourceBlock && !logs.some((d) => block.permutation.matches(d)))
    return;

  const column = locationObject[block.x]?.[block.z];

  if (column && column.max >= block.y && column.min <= block.y) return;

  locationObject[block.x] ??= {};
  locationObject[block.x][block.z] ??= { min: block.y, max: block.y };

  const object = locationObject[block.x][block.z];

  object.min = Math.min(object.min, block.y);
  object.max = Math.max(object.max, block.y);

  await Promise.all([
    !bannedDirections.includes("above") &&
      fellBlock(
        [...bannedDirections, "below"],
        block.above(),
        sourceBlock,
        locationObject
      ),
    !bannedDirections.includes("below") &&
      fellBlock(
        [...bannedDirections, "above"],
        block.below(),
        sourceBlock,
        locationObject
      ),
  ]);

  await Promise.all([
    !bannedDirections.includes("north") &&
      fellBlock(
        [...bannedDirections, "south"],
        block.north(),
        sourceBlock,
        locationObject
      ),
    !bannedDirections.includes("east") &&
      fellBlock(
        [...bannedDirections, "west"],
        block.east(),
        sourceBlock,
        locationObject
      ),
    !bannedDirections.includes("south") &&
      fellBlock(
        [...bannedDirections, "north"],
        block.south(),
        sourceBlock,
        locationObject
      ),
    !bannedDirections.includes("west") &&
      fellBlock(
        [...bannedDirections, "east"],
        block.west(),
        sourceBlock,
        locationObject
      ),
  ]);

  return block.location;
}

export async function handleFellTree(sourceBlock) {
  try {
    const locationObject = {};

    const location = await fellBlock(
      [],
      sourceBlock,
      sourceBlock,
      locationObject
    );
    if (!location) return;

    world.playSound("mob.zombie.woodbreak", location);

    system.run(() =>
      sourceBlock.dimension.runCommandAsync(
        `setblock ${location.x} ${location.y} ${location.z} air replace`
      )
    );

    for (const [x, value] of Object.entries(locationObject)) {
      const columns = Object.entries(value);
      for (const [z, { min, max }] of columns) {
        system.runTimeout(
          () =>
            sourceBlock.dimension.runCommandAsync(
              `fill ${x} ${min} ${z} ${x} ${max} ${z} air destroy`
            ),
          Math.random() * columns.length * 2
        );
      }
    }
  } catch (e) {
    log(`Error in tree felling: ${e}`);
  }
}
