import { EntityHealthComponent, system } from "@minecraft/server";
import {
  addExp,
  getAbilityName,
  getLevel,
  getSelectedAbility,
  getSkillName,
} from "../../utils/addonUtils";
import TimerUtil, { AbilityTimerUtil } from "../../utils/cooldowns";
import { triggerAbilityEvent } from "../../utils/events";
import { displayActionBar, getDistance, isHolding } from "../../utils/utils";
import { triple_shot_ability } from "../abilities/archer.triple_shot";
import { arrowRegistry } from "../registry";
import { explosive_arrow_ability } from "../abilities/archer.explosive_arrow";
import { arrow_shield_ability } from "../abilities/archer.arrow_shield";
import { rewardKill } from "./fighting";
import { showSkillInfoPage } from "../../ui";

const abilityCharge = new TimerUtil();

export const archer_skill = {
  key: "archer",
  primary_color: "§c",
  secondary_color: "§n",
  abilityCooldown: new AbilityTimerUtil(),
  base_exp: 30,
  base_passive_chance: 1,
  entityHurt: {
    onEvent: async (event) => {
      const { damageSource, hurtEntity, damage } = event;
      if (
        event.damageSource.damagingEntity?.typeId === "minecraft:player" &&
        event.damageSource.cause === "projectile"
      ) {
        const level = getLevel(event.damageSource.damagingEntity, archer_skill);
        const component = hurtEntity.getComponent(
          EntityHealthComponent.componentId
        );
        component.setCurrentValue(
          component.currentValue -
            damage * ((archer_skill.base_passive_chance * level) / 100)
        );
        if (
          hurtEntity.matches({ families: ["monster"] }) &&
          component.currentValue <= 0
        )
          rewardKill(damageSource.damagingEntity, hurtEntity);
      }
    },
  },
  projectileHit: {
    block: {
      onEvent: async (event) => handleProjectileHit(event, "block"),
    },
    entity: {
      onEvent: async (event) => handleProjectileHit(event, "entity"),
    },
  },
  playerShootArrow: {
    onEvent: async (event) => handleShootArrow(event),
  },
  playerSneak: {
    onEvent: (event, doubleClick) => {
      if (
        !event.isSneaking ||
        !(
          isHolding(event.player, "minecraft:bow") ||
          isHolding(event.player, "minecraft:arrow")
        )
      )
        return;
      abilityActionBar(event.player, doubleClick, "Sneak");
      event.player.playSound("random.click");
      if (
        getSelectedAbility(event.player, archer_skill)?.key === "arrow_shield"
      ) {
        if (doubleClick)
          triggerAbilityEvent.call({
            player: event.player,
            skill: archer_skill,
            eventData: event,
          });
      } else {
        if (doubleClick) chargeAbility(event.player);
      }
    },
  },
  itemUse: {
    item: {
      typeIds: ["minecraft:arrow"],
    },
    double_click: {
      useCommon: false,
      sneak: {
        onEvent: (event, doubleClick) => {
          event.source.playSound("random.click");
          displayActionBar(event.source, 1, {
            rawtext: [
              ...getSkillName(archer_skill, archer_skill.primary_color),
              {
                text: ` §8- §aClick §7- ${doubleClick ? "§a" : "§7"}Click §8- `,
              },
              { translate: "pod_rpg.common.open_menu" },
            ],
          });
          if (doubleClick) showSkillInfoPage(event.source, archer_skill);
        },
      },
      normal: {
        onEvent: (event, doubleClick) => {
          event.source.playSound("random.click");
          abilityActionBar(event.source, doubleClick, "Click");
          if (doubleClick) chargeAbility(event.source);
        },
      },
    },
  },
  abilities: [
    triple_shot_ability,
    explosive_arrow_ability,
    arrow_shield_ability,
  ],
};

const handleProjectileHit = async (event, type) => {
  if (event.projectile.typeId !== "minecraft:arrow") return;
  const arrow = arrowRegistry[event.projectile.id];
  if (type === "entity") {
    const source = event.source
      ? event.source
      : arrow
      ? arrow.source
      : undefined;
    if (source?.typeId !== "minecraft:player") return;
    if (source)
      addExp(
        source,
        archer_skill,
        Math.ceil(getDistance(source.location, event.location) / 4)
      );
  }
};

const handleShootArrow = async (event) => {
  const { player } = event;
  const ability = getSelectedAbility(player, archer_skill);
  addExp(player, archer_skill, Math.ceil(Math.min(20, event.useDuration) / 4));

  if (
    abilityCharge.has(player.name) &&
    (ability?.key === "triple_shot" || ability?.key === "explosive_arrow")
  ) {
    abilityCharge.set(player.name, 0);
    ability.onEnable(event);
  }
};

export const chargeAbility = (player) => {
  const ability = getSelectedAbility(player, archer_skill);

  if (
    ability === undefined ||
    archer_skill.abilityCooldown.has(player.id + ability.key)
  )
    return;
  if (ability.key === "arrow_shield") {
    triggerAbilityEvent.call({
      player: player,
      skill: archer_skill,
    });
    return;
  }
  const charged = abilityCharge.has(player.name);
  abilityCharge.set(player.name, charged ? 0 : 60 * 20);
  if (!charged)
    abilityCharge.setOnExpire(player.name, () => {
      player.playSound("armor.equip_generic", { pitch: 0.7 });
      displayActionBar(player, 0, {
        rawtext: [
          ...getAbilityName(ability, archer_skill.secondary_color),
          { text: " " },
          { translate: "pod_rpg.skills.archer.charge_expire" },
        ],
      });
    });
  player.playSound("armor.equip_generic", { pitch: charged ? 0.7 : 1 });
  displayActionBar(player, 0, {
    rawtext: [
      ...getAbilityName(ability, archer_skill.secondary_color),
      { text: " " },
      {
        translate: charged
          ? "pod_rpg.skills.archer.charge_expire"
          : "pod_rpg.skills.archer.charge",
      },
    ],
  });
};

const abilityActionBar = (player, doubleClick, actionKey) => {
  const ability = getSelectedAbility(player, archer_skill);
  displayActionBar(player, 1, {
    rawtext: [
      ...(ability === undefined
        ? getSkillName(archer_skill, archer_skill.primary_color)
        : getAbilityName(ability, archer_skill.secondary_color)),
      {
        text: ` §8- §a${actionKey} §7- ${
          doubleClick ? "§a" : "§7"
        }${actionKey} §8- `,
      },
      ...getAbilityStatus(player, ability),
    ],
  });
};

const getAbilityStatus = (player, ability) => {
  if (ability === undefined)
    return [
      { text: "§c" },
      { translate: "pod_rpg.common.ability.non_selected" },
    ];
  if (ability && !ability.onEnable) {
    return [
      { text: "§p" },
      { translate: "pod_rpg.common.ability.always_active" },
    ];
  }

  if (archer_skill.abilityCooldown.has(player.id + ability.key))
    return [
      { text: `§c` },
      {
        translate: "pod_rpg.common.ability.cooldown",
        with: [
          (
            (archer_skill.abilityCooldown.get(player.id + ability.key) -
              system.currentTick) /
            20
          ).toFixed(1),
        ],
      },
    ];
  if (ability.key === "arrow_shield") {
    return [{ text: "§a" }, { translate: "pod_rpg.common.ability.ready" }];
  } else {
    return abilityCharge.has(player.name)
      ? [{ text: "§a" }, { translate: "pod_rpg.skills.archer.is_charged" }]
      : [{ text: "§c" }, { translate: "pod_rpg.skills.archer.is_not_charged" }];
  }
};
