import { EquipmentSlot, system } from "@minecraft/server";
import {
  abilityActionBar,
  addExp,
  dropAdditionalItem,
  getAbilityLevel,
  getLevel,
  getSelectedAbility,
} from "../../utils/addonUtils";
import {
  addCoords,
  blockMatches,
  getItemInSlot,
  permutationMatches,
} from "../../utils/utils";
import { AbilityTimerUtil } from "../../utils/cooldowns";
import { pod_farm_crop_definitions } from "../pod_farm";
import { isAddonInstalled } from "../../main";
import { farm_exp_ability } from "../abilities/farming.farm_exp";
import {
  handleHarvestAbility,
  harvest_ability,
  harvest_timer,
} from "../abilities/farming.harvest";
import {
  fertiliser_ability,
  fertiliser_timer,
  handleFertiliserAbility,
} from "../abilities/farming.fertiliser";
import { triggerAbilityEvent } from "../../utils/events";

export const farming_skill = {
  key: "farming",
  primary_color: "§g",
  secondary_color: "§p",
  abilityCooldown: new AbilityTimerUtil(),
  base_exp: 30,
  base_passive_chance: 0.5,
  playerBreakBlock: {
    disallowPlayerPlacedBlocks: false,
    onEvent: (event) => handleBlockBreak(event),
  },
  entityHitEntity: {
    onEvent: async (event) => handleEntityHit(event),
  },
  dataDrivenEntityTrigger: {
    onEvent: async (event) => handleEntityTrigger(event),
  },
  playerSneak: {
    onEvent: (event, doubleClick) => {
      if (
        !event.isSneaking ||
        !getItemInSlot(event.player, EquipmentSlot.Mainhand)?.hasTag(
          "minecraft:is_hoe"
        )
      )
        return;
      abilityActionBar(event.player, farming_skill, doubleClick, "Sneak");
      event.player.playSound("random.click");
      if (doubleClick) {
        triggerAbilityEvent.call({
          player: event.player,
          skill: farming_skill,
          eventData: event,
        });
      }
    },
  },
  itemUseOn: {
    after: {
      onEvent: async (event) => handleItemUseOnAfter(event),
    },
    before: {
      onEvent: async (event) => {
        const reward = farming_rewards[0];
        if (blockMatches(event.block, reward.blocks))
          system.runTimeout(() =>
            rewardCrop(event.source, event.block, reward)
          );
      },
    },
  },
  itemUse: {
    item: {
      tags: ["minecraft:is_hoe"],
    },
    double_click: {
      useCommon: true,
    },
  },
  abilities: [farm_exp_ability, harvest_ability, fertiliser_ability],
};

const handleItemUseOnAfter = async (event) => {
  const { itemStack, block } = event;
  if (
    itemStack.typeId === "minecraft:bone_meal" &&
    fertiliser_timer.has(event.source.name)
  ) {
    handleFertiliserAbility(
      block.dimension,
      addCoords(block.location, { x: 0.5, y: 0.5, z: 0.5 })
    );
  }
};

const handleEntityTrigger = async (event) => {
  //POD FARM INTEGRATION
  if (!isAddonInstalled("pod_farm")) return;
  const { entity, eventId } = event;
  let player;
  switch (eventId) {
    case "pod_farm:bone_mealed":
      player = entity.dimension.getEntities({
        closest: 1,
        location: entity.location,
        maxDistance: 5,
        type: "minecraft:player",
      })[0];
      if (fertiliser_timer.has(player.name))
        handleFertiliserAbility(entity.dimension, entity.location);
      break;
    case "pod_farm:drop":
      if (entity.matches({ families: ["pod_farm_crop"] })) {
        const crop = pod_farm_crop_definitions.find(
          (v) => v.id === entity.typeId.replace("pod_farm:", "")
        );
        if (!crop) return;
        const player = entity.dimension.getEntities({
          closest: 1,
          location: entity.location,
          maxDistance: 5,
          type: "minecraft:player",
        })[0];
        if (!player) return;
        rewardCrop(
          player,
          addCoords(entity.location, { x: 0, y: 0.5, z: 0 }),
          crop
        );
      }
      break;
  }
};

const handleEntityHit = async (event) => {
  //POD FARM INTEGRATION
  if (!isAddonInstalled("pod_farm")) return;
  const { hitEntity, damagingEntity } = event;
  if (damagingEntity.typeId !== "minecraft:player") return;
  if (
    event.hitEntity?.isValid() &&
    event.hitEntity.matches({ families: ["pod_farm_crop"] })
  ) {
    //HARVEST ABILITY
    if (harvest_timer.has(damagingEntity.name)) {
      handleHarvestAbility(hitEntity.dimension, hitEntity.location);
    }

    const crop = pod_farm_crop_definitions.find(
      (v) => v.id === hitEntity.typeId.replace("pod_farm:", "")
    );
    if (!crop || hitEntity.getProperty("pod_farm:growth") !== crop.max_growth)
      return;
    rewardCrop(
      damagingEntity,
      addCoords(hitEntity.location, { x: 0, y: 0.5, z: 0 }),
      crop
    );
  }
};

const handleBlockBreak = (event) => {
  const { brokenBlockPermutation, player, block } = event;

  //HARVEST ABILITY
  if (harvest_timer.has(player.name)) {
    handleHarvestAbility(
      block.dimension,
      addCoords(block.location, { x: 0.5, y: 0.5, z: 0.5 })
    );
  }

  for (const reward of farming_rewards) {
    if (permutationMatches(brokenBlockPermutation, reward.blocks)) {
      rewardCrop(
        player,
        addCoords(block.location, { x: 0.5, y: 0.5, z: 0.5 }),
        reward
      );
      break;
    }
  }
};

export const rewardCrop = (player, location, reward) => {
  //EXP REWARD
  addExp(player, farming_skill, reward.exp);

  //EXPERIENCE ABILITY
  const selectedAbility = getSelectedAbility(player, farming_skill);
  if (selectedAbility && selectedAbility.key === "farm_exp") {
    const abilityLevel = getAbilityLevel(
      player,
      farming_skill,
      selectedAbility
    );
    if (selectedAbility.chance(abilityLevel) >= Math.random() * 100) {
      for (let i = 0; i < (abilityLevel >= 4 ? 2 : 1); i++) {
        player.dimension.spawnEntity("minecraft:xp_orb", location);
      }
    }
  }

  //ITEM DROP
  if (!reward.item) return;
  const chance =
    getLevel(player, farming_skill) * farming_skill.base_passive_chance;
  dropAdditionalItem(
    location,
    player.dimension,
    chance,
    reward.item,
    reward.item_count || 1
  );
};

export const farming_rewards = [
  //Keep sweet berry bush first, (see itemUseOn.b.w.js)
  {
    blocks: [{ block: "sweet_berry_bush", states: { growth: 3 } }],
    item: "minecraft:sweet_berries",
    exp: 5,
  },
  {
    blocks: [{ block: "wheat", states: { growth: 7 } }],
    item: "minecraft:wheat",
    exp: 5,
  },
  {
    blocks: [{ block: "potatoes", states: { growth: 7 } }],
    item: "minecraft:potato",
    exp: 5,
  },
  {
    blocks: [{ block: "carrots", states: { growth: 7 } }],
    item: "minecraft:carrot",
    exp: 5,
  },
  {
    blocks: ["melon_block"],
    item: "minecraft:melon_slice",
    item_count: 5,
    exp: 5,
  },
  {
    blocks: ["pumpkin"],
    item: "minecraft:pumpkin",
    exp: 5,
  },
  {
    blocks: [{ block: "cocoa", states: { age: 2 } }],
    item: "minecraft:cocoa_beans",
    exp: 5,
  },
  {
    blocks: [{ block: "nether_wart", states: { age: 3 } }],
    item: "minecraft:nether_wart",
    exp: 5,
  },
  {
    blocks: [{ block: "beetroot", states: { growth: 7 } }],
    item: "minecraft:beetroot",
    exp: 5,
  },
];
