import { system, world } from "@minecraft/server";
import { blockMatches, log } from "./utils/utils";

import "./utils/events";
import "./events/playerBreakBlock.a.w";
import "./events/playerSpawn.a.w";
import "./events/itemUse.a.w";
import "./events/scriptEventReceive.a.s";
import "./events/playerPlaceBlock.a.w";
import "./events/entitySpawn.a.w";
import "./events/projectileHitBlock.a.w";
import "./events/projectileHitEntity.a.w";
import "./events/custom/playerSneakEvent";
import "./events/custom/triggerAbilityEvent";
import { mining_skill } from "./data/skills/mining";
import { tree_cutting_skill } from "./data/skills/tree_cutting";
import { farming_skill } from "./data/skills/farming";
import { fighting_skill } from "./data/skills/fighting";
import { blastmaster_skill } from "./data/skills/blastmaster";
import { explorer_skill } from "./data/skills/explorer";
import { pyromaniac_skill } from "./data/skills/pyromaniac";
import { getSetting, loadSettings } from "./utils/settings";
import {
  playerJumpEvent,
  playerShootArrowEvent,
  playerSneakEvent,
} from "./utils/events";
import { archer_skill } from "./data/skills/archer";
import { animals, hunting_skill } from "./data/skills/hunting";
import { pod_farm_animal_definitions } from "./data/pod_farm";

export const skills = [
  mining_skill,
  tree_cutting_skill,
  farming_skill,
  fighting_skill,
  blastmaster_skill,
  explorer_skill,
  pyromaniac_skill,
  archer_skill,
  hunting_skill,
];

export const dimensions = [
  world.getDimension("overworld"),
  world.getDimension("nether"),
  world.getDimension("the_end"),
];

export const integratedAddons = {
  pod_farm: false,
};

export const isAddonInstalled = (addon_id) => {
  if (getSetting("addon_integration")?.valueOf() === false) return false;
  return integratedAddons[addon_id];
};

const playerSneakData = {};
const playerJumpData = {};

export async function eventHandler() {
  for (const player of world.getAllPlayers()) {
    if (
      playerSneakData[player.name] === undefined ||
      playerSneakData[player.name] !== player.isSneaking
    ) {
      playerSneakEvent.call({
        player: player,
        isSneaking: player.isSneaking,
        tick: system.currentTick,
      });

      playerSneakData[player.name] = player.isSneaking;
    }

    if (system.currentTick % 1 === 0) {
      const jumpData = playerJumpData[player.name];
      if (!jumpData || jumpData.jumping !== player.isJumping) {
        playerJumpEvent.call({
          player: player,
          isJumping: player.isJumping,
          lastJump: jumpData?.lastJump,
          tick: system.currentTick,
        });

        playerJumpData[player.name] = {
          jumping: player.isJumping,
          lastJump: player.isJumping
            ? system.currentTick
            : jumpData?.lastJump || 0,
        };
      }
    }
  }
}

//SETUP
system.runTimeout(async () => {
  await setup();
  system.runInterval(async () => eventHandler());
  system.runInterval(async () => removeAnimalRadar(), 20);
  system.runInterval(async () => handlePlayerEffects(), 10);
}, 100);

const removeAnimalRadar = async () => {
  for (const entity of world
    .getDimension("overworld")
    .getEntities({ type: "pod_rpg:animal_radar" })) {
    if (
      entity.dimension.getPlayers({
        location: entity.location,
        maxDistance: 30,
      }).length === 0
    )
      entity.remove();
  }
};

const handlePlayerEffects = async () => {
  for (const player of world.getPlayers({ tags: ["pod_rpg.slowFalling"] })) {
    if (player.isOnGround || player.isGliding) {
      player.removeEffect("slow_falling");
      player.removeTag("pod_rpg.slowFalling");
    }
  }
  for (const player of world.getPlayers({
    tags: ["pod_rpg.fireResistance"],
  })) {
    if (
      !blockMatches(player.dimension.getBlock(player.location).below(), [
        "magma",
        "lava",
        "flowing_lava",
      ])
    ) {
      player.removeEffect("fire_resistance");
      player.removeTag("pod_rpg.fireResistance");
    }
  }
};

const setup = async () => {
  log("Setup");
  loadSettings();
  let isPodFarmInstalled = false;
  try {
    isPodFarmInstalled =
      world
        .getDimension("overworld")
        .runCommand('loot spawn ~ -100 ~ loot "pod_farm_animals/bison"')
        .successCount === 1;
  } catch (e) {
    log("Couldn't validate if pod_farm is installed");
  }
  integratedAddons["pod_farm"] = isPodFarmInstalled;

  if (isAddonInstalled("pod_farm"))
    animals.push(...pod_farm_animal_definitions);

  //Skill Events
  for (const skill of skills) {
    if (skill.tick)
      system.runInterval(async () => skill.tick.onTick(), skill.tick.interval);
    if (skill.playerJump)
      playerJumpEvent.subscribe(async (event) =>
        skill.playerJump.onEvent(event)
      );
    if (skill.playerShootArrow)
      playerShootArrowEvent.subscribe(async (event) =>
        skill.playerShootArrow.onEvent(event)
      );
    if (skill.dataDrivenEntityTrigger)
      world.afterEvents.dataDrivenEntityTrigger.subscribe(async (event) =>
        skill.dataDrivenEntityTrigger.onEvent(event)
      );
    if (skill.entityDie)
      world.afterEvents.entityDie.subscribe(async (event) =>
        skill.entityDie.onEvent(event)
      );
    if (skill.entityHitEntity)
      world.afterEvents.entityHitEntity.subscribe(async (event) =>
        skill.entityHitEntity.onEvent(event)
      );
    if (skill.entityHurt)
      world.afterEvents.entityHurt.subscribe(async (event) =>
        skill.entityHurt.onEvent(event)
      );
    if (skill.itemUseOn?.after)
      world.afterEvents.itemUseOn.subscribe(async (event) =>
        skill.itemUseOn.after.onEvent(event)
      );
    if (skill.itemUseOn?.before)
      world.beforeEvents.itemUseOn.subscribe(async (event) =>
        skill.itemUseOn.before.onEvent(event)
      );
    if (skill.projectileHit?.block)
      world.afterEvents.projectileHitBlock.subscribe(async (event) =>
        skill.projectileHit.block.onEvent(event)
      );
    if (skill.projectileHit?.entity)
      world.afterEvents.projectileHitEntity.subscribe(async (event) =>
        skill.projectileHit.entity.onEvent(event)
      );
  }
};
