import { ItemStack, system, world } from "@minecraft/server";
import { showConfirmationScreen } from "./uiUtils";
import { displayActionBar, displayTitle, spawnParticle } from "./utils";
import { getSetting } from "./settings";
import { MessageFormData } from "@minecraft/server-ui";

export const MAX_LEVEL = 100;

export const setSelectedAbility = (player, skill, ability) => {
  player.setDynamicProperty(`pod_rpg_${skill.key}_ability`, ability);
};

export const getSelectedAbility = (player, skill) => {
  const i = player.getDynamicProperty(`pod_rpg_${skill.key}_ability`);
  if (i === undefined || i < 0) return undefined;
  return skill.abilities[i];
};

export const addExp = (player, skill, amount, priority = 3) => {
  if (amount === 0) return;
  const exp =
    (player.getDynamicProperty(`pod_rpg_${skill.key}_exp`) || 0) + amount;
  player.setDynamicProperty(`pod_rpg_${skill.key}_exp`, exp);
  const level = getLevelFromExp(skill, exp);
  const prevLevel = getLevelFromExp(skill, exp - amount);
  //LEVEL CAP 100
  if (prevLevel === MAX_LEVEL) return;
  const expForThisLevel = getExpRequiredForLevel(skill, level);
  const expRequired =
    getExpRequiredForLevel(skill, level + 1) - expForThisLevel;

  if (
    skill.key !== "explorer" ||
    getSetting("explorer_exp_bar")?.valueOf() === true
  ) {
    displayActionBar(player, priority, {
      rawtext: [
        ...getSkillName(skill, skill.primary_color),
        {
          text: `  §8- [${getRenderableExpBar(
            exp - expForThisLevel,
            expRequired,
            30
          )}§8] - §gLv. ${level}`,
        },
      ],
    });
  }

  if (level > prevLevel) {
    if (
      getSetting("announce_level_up")?.valueOf() === true &&
      level % 5 === 0
    ) {
      world.sendMessage({
        rawtext: [
          { text: `§7[` },
          ...getSkillName(skill, skill.primary_color),
          { text: `§7] §r` },
          {
            translate: "pod_rpg.common.levelup",
            with: [player.name, `${level}`],
          },
        ],
      });
    }

    if (getSetting("ability_unlock_message")?.valueOf() === true) {
      for (const ability of skill.abilities) {
        if (ability.unlock_level > prevLevel && ability.unlock_level <= level) {
          player.sendMessage({
            rawtext: [
              { text: `§7[` },
              ...getSkillName(skill, skill.primary_color),
              { text: `§7] ${skill.secondary_color}` },
              ...getAbilityName(ability),
              { text: " §r" },
              { translate: "pod_rpg.common.ability.unlock" },
            ],
          });
        }
      }
    }

    player.playSound("random.levelup");
    spawnParticle(player.dimension, player.location, "pod_rpg:levelup");
    displayTitle(
      player,
      {
        rawtext: getSkillName(skill, skill.primary_color),
      },
      {
        rawtext: [{ text: `§f${prevLevel} §7-> §f${level}` }],
      }
    );
  }
};

export const abilityActionBar = (player, skill, doubleClick, actionKey) => {
  const ability = getSelectedAbility(player, skill);
  displayActionBar(player, 1, {
    rawtext: [
      ...(ability === undefined
        ? getSkillName(skill, skill.primary_color)
        : getAbilityName(ability, skill.secondary_color)),
      {
        text: ` §8- §a${actionKey} §7- ${
          doubleClick ? "§a" : "§7"
        }${actionKey} §8- `,
      },
      ...getAbilityStatus(player, skill, ability),
    ],
  });
};

const getAbilityStatus = (player, skill, ability) => {
  if (ability === undefined)
    return [
      { text: "§c" },
      { translate: "pod_rpg.common.ability.non_selected" },
    ];
  if (ability && !ability.onEnable) {
    return [
      { text: "§p" },
      { translate: "pod_rpg.common.ability.always_active" },
    ];
  }

  if (skill.abilityCooldown.has(player.id + ability.key))
    return [
      { text: `§c` },
      {
        translate: "pod_rpg.common.ability.cooldown",
        with: [
          (
            (skill.abilityCooldown.get(player.id + ability.key) -
              system.currentTick) /
            20
          ).toFixed(1),
        ],
      },
    ];
  return [{ text: "§a" }, { translate: "pod_rpg.common.ability.ready" }];
};

export const getRenderableExpBar = (exp, expRequired, length) => {
  const expBarGreen = Math.ceil((exp / expRequired) * length);
  return (
    "§a" +
    "|".repeat(length).substring(0, expBarGreen) +
    "§7" +
    "|".repeat(length).substring(expBarGreen)
  );
};

export const getLevelFromExp = (skill, exp) => {
  return Math.min(
    Math.floor((-1 + Math.sqrt(1 + 8 * (exp / skill.base_exp))) / 2),
    MAX_LEVEL
  );
};

export const getLevel = (player, skill) => {
  return getLevelFromExp(
    skill,
    player.getDynamicProperty(`pod_rpg_${skill.key}_exp`) || 0
  );
};

export const getExpRequiredForLevel = (skill, level) => {
  return (level / 2) * skill.base_exp * (level + 1);
};

export const enableAbilityTitle = (player, ability, color) => {
  displayTitle(player, "§r", {
    rawtext: [
      { text: color },
      ...getAbilityName(ability),
      { text: " " },
      {
        translate: "pod_rpg.common.ability.enable",
      },
    ],
  });
};

export const getAbilityName = (ability, formatting = "") => {
  return [
    { text: formatting },
    { translate: `pod_rpg.abilities.${ability.key}.name` },
  ];
};

export const getSkillName = (skill, formatting = "") => {
  return [
    { text: formatting },
    { translate: `pod_rpg.skills.${skill.key}.name` },
  ];
};

export const getAbilityLevel = (player, skill, ability) => {
  if (Number.isInteger(ability)) ability = skill.abilities[ability];
  return player.getDynamicProperty(`pod_rpg_${skill.key}_${ability.key}`) || 0;
};

export const dropAdditionalItem = (
  location,
  dimension,
  chance,
  item,
  item_count
) => {
  if (chance >= Math.random() * 100) {
    const spawnedItem = dimension.spawnItem(
      new ItemStack(item, item_count),
      location
    );
    spawnParticle(
      spawnedItem.dimension,
      spawnedItem.location,
      "pod_rpg:extra_drop"
    );
    world.playSound("random.pop", spawnedItem.location);
  }
};

export const showResetUi = (player) => {
  const form = new MessageFormData();
  form.title({ rawtext: [{ translate: "pod_rpg.reset.title1" }] });
  form.body({ rawtext: [{ translate: "pod_rpg.reset.disclaimer1" }] });
  form.button1({
    rawtext: [{ translate: "pod_rpg.common.ui.cancel_button" }],
  });
  form.button2({
    rawtext: [{ translate: "pod_rpg.common.ui.continue_button" }],
  });
  showConfirmationScreen(player, form, () => {
    form.title({ rawtext: [{ translate: "pod_rpg.reset.title2" }] });
    form.body({ rawtext: [{ translate: "pod_rpg.reset.disclaimer2" }] });
    form.button2({ rawtext: [{ translate: "pod_rpg.reset.button" }] });
    showConfirmationScreen(player, form, () => player.clearDynamicProperties());
  });
};
