import {
  EntityEquippableComponent,
  EntityHealthComponent,
  EntityInventoryComponent,
  EquipmentSlot,
  system,
  world,
} from "@minecraft/server";
import { getSetting } from "./settings";

// Overrides and monkey patching

export const handleAsync = (promise, where = "") => {
  promise.catch((e) => {
    log(`Error occurred ${where.length ? `[${where}]` : ""} : §c${e}`);
  });
};

export function getDistance(a, b) {
  const x = a.x - b.x;
  const y = a.y - b.y;
  const z = a.z - b.z;
  return Math.sqrt(x * x + y * y + z * z);
}

export function getMagnitude(v) {
  return Math.sqrt(v.x * v.x + v.y * v.y + v.z * v.z);
}

export function getAngle(a, b) {
  const dx = b.x - a.x;
  const dz = b.z - a.z;
  const theta = Math.atan2(dz, dx) * (180 / Math.PI);
  return theta < 0 ? 360 + theta : theta;
}

export function getDirection(a, b) {
  return { x: b.x - a.x, y: b.y - a.y, z: b.z - a.z };
}

export function rotateVector(vector, angle) {
  angle = -angle * (Math.PI / 180);
  var cos = Math.cos(angle);
  var sin = Math.sin(angle);
  return {
    x: vector.x * cos - vector.z * sin,
    y: vector.y,
    z: vector.x * sin + vector.z * cos,
  };
}

export const addCoords = (a, b) => {
  return {
    x: a.x + (b.x || 0),
    y: a.y + (b.y || 0),
    z: a.z + (b.z || 0),
  };
};

export function getItemInSlot(player, slot) {
  const component = player.getComponent(EntityEquippableComponent.componentId);
  return component?.getEquipment(slot);
}

export function isHolding(player, item) {
  return getItemInSlot(player, EquipmentSlot.Mainhand)?.typeId === item;
}

export const isAlive = (entity) => {
  if (
    !entity.hasComponent(EntityHealthComponent.componentId) ||
    entity.getComponent(EntityHealthComponent.componentId).currentValue <= 0
  )
    return false;
  return true;
};

const currentBar = {
  priority: 1000,
  since: 0,
};
export const displayActionBar = (player, priority, text) => {
  if (getSetting("reduce_validations")?.valueOf() === true) {
    player.onScreenDisplay.setActionBar(text);
    return;
  }

  currentBar.priority += Math.max(0, system.currentTick - currentBar.since - 2);
  if (priority <= currentBar.priority) {
    player.onScreenDisplay.setActionBar(text);
    currentBar.priority = priority;
    currentBar.since = system.currentTick;
  }
};

export const isString = (value) => {
  return typeof value === "string" || value instanceof String;
};

export const hasItem = (player, typeId, amount) => {
  const container = player.getComponent(
    EntityInventoryComponent.componentId
  )?.container;
  if (!container || container.emptySlotsCount === container.size) return false;
  let found = 0;
  for (let i = 0; i < container.size; i++) {
    const item = container.getItem(i);
    if (!item || !item.typeId || !item.amount) continue;
    if (item.typeId === typeId) {
      if (item.amount + found >= amount) return true;
      found += item.amount;
    }
  }
  return false;
};

export const permutationMatches = (permutation, permutationList) => {
  return indexOfPermutationMatched(permutation, permutationList) !== -1;
};

export const indexOfPermutationMatched = (permutation, permutationList) => {
  if (isString(permutationList[0])) {
    permutationList = permutationList.map((v) => {
      return { block: v, states: {} };
    });
  }
  for (const toCheck of permutationList) {
    if (permutation?.matches(toCheck.block, toCheck.states) ?? true) {
      return permutationList.indexOf(toCheck);
    }
  }
  return -1;
};

export const blockMatches = (block, blockIds) => {
  return permutationMatches(block.permutation, blockIds);
};

export const spawnParticle = (dimension, location, particleName) => {
  if (getSetting("particles")?.valueOf() === false) return;
  dimension.spawnParticle(particleName, location);
};

export const displayTitle = (
  player,
  title = "",
  subtitle = "",
  options = undefined
) => {
  player.onScreenDisplay.updateSubtitle(subtitle);
  if (options === undefined) player.onScreenDisplay.setTitle(title);
  else player.onScreenDisplay.setTitle(title, options);
};

export const runCmdSync = async (commandString) =>
  world.getDimension("overworld").runCommand(commandString);

const overworld = world.getDimension("overworld");

export const locationString = (location) => {
  return `${location.x} ${location.y} ${location.z}`;
};

function processObject(object) {
  if (!Object.keys(object).length) return "§7§oEmpty Object§r";

  let output = "{ ";

  for (const [key, value] of Object.entries(object)) {
    output += `§6${key}§7: §r${process(value)} `;
  }
  return `${output}}`;
}

function process(arg) {
  let output = arg?.toString();

  if (Array.isArray(arg)) output = `[${arg.map((d) => process(d)).join(", ")}]`;
  else if (arg === null) output = `§6null`;
  else
    switch (typeof arg) {
      case "object":
        output = processObject(arg);
        break;
      case "string":
        output = `${arg.startsWith("Error") ? "§c" : '§a"§2'}${arg}§a${
          arg.startsWith("Error") ? "" : '"'
        }`;
        break;
      case "number":
        output = isNaN(arg) ? `§5§oNaN` : `§9${arg}`;
        break;
      case "boolean":
        output = `§6${output}`;
        break;
      case "undefined":
        output = `§6undefined`;
        break;
    }

  return `${output}§r`;
}

/**
 * Formats arguments with colour and outputs them to Minecraft chat via tellraw.
 *
 * @param args
 */
export function log(...args) {
  overworld
    .runCommandAsync(
      `tellraw @a[tag=pod_rpg_receive_logs] {"rawtext":[{"text":"${args
        .map((d) => process(d))
        .join(" ")
        .replaceAll('"', '\\"')}"}]}`
    )
    .then()
    .catch();
}

export function warn(...args) {
  // eslint-disable-next-line no-undef
  console.warn(...args.map(process));
}

export function error(...args) {
  // eslint-disable-next-line no-undef
  console.error(...args.map(process));
}

export const runCmd = async (commandString) =>
  await world.getDimension("overworld").runCommandAsync(commandString);
